/*
 * Tiny TTY driver
 *
 * Copyright (C) 2002 Greg Kroah-Hartman (greg@kroah.com)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; version 2 of the License.
 *
 * This driver shows how to create a minimal serial driver using the 2.5
 * kernel's serial driver layer. It does not rely on any backing hardware, but
 * creates a timer that emulates data being received from some kind of
 * hardware.
 *
 * This code is based on the exar driver
 *
 * Compiled this driver with:
 *
 *     CFLAGS = -Wall -O2 -fomit-frame-pointer -DMODULE -D__KERNEL__
 *     IDIR = /path/to/linux-2.5/include/goes/here
 *
 *     tiny_serial.o::tiny_serial.c
 *     $(CC) $(CFLAGS) -I$(IDIR) -c -o $@ $<
 */

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/ioport.h>
#include <linux/init.h>
#include <linux/console.h>
#include <linux/sysrq.h>
#include <linux/mca.h>
#include <linux/delay.h>
#include <linux/platform_device.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/serial_reg.h>
#include <linux/serial_core.h>
#include <linux/serial.h>
#include <linux/serial_8250.h>
#include <linux/nmi.h>
#include <linux/notifier.h>
#include <linux/reboot.h>
#include <asm/io.h>
#include <asm/irq.h>

#include <asm-mips/mach-vsc452/vsc452.h>
#include <asm/addrspace.h>

#include "expander.h"
#define EXPANDER_UART_ID	0

/**********************************************
*   Code Configuration
*
***********************************************/
#define SUPPORT_EXPANDER_UART_TASKLET 1
#define ENABLE_EXAR_INTERRUPT       /* If defined, use interrupt-driven RX and TX, otherwise poll */
//#define CONFIG_EXAR_16M770_CONSOLE  /* If defined, select EXAR_CONSOLE */
//#define EXAR_DEBUG                  /* If defined, enable DEBUG for general functions */
//#define EXAR_INTR                   /* If defined, enable DEBUG in ISR, RX and TX functions */
//#define EXAR_IO                     /* If defined, output minimal information for RX and TX bytes */

/* If not using interrupts, poll for RX and TX */
#ifndef ENABLE_EXAR_INTERRUPT
#define ENABLE_EXAR_POLLING  /* Use polling for RX and TX */
#endif

#ifdef EXAR_DEBUG
# define DEBUG_PRINT(...) printk(__VA_ARGS__)
#else
# define DEBUG_PRINT(...)
#endif

#ifdef EXAR_INTR
# define DEBUG_INTR(...) printk(__VA_ARGS__)
#else
# define DEBUG_INTR(...)
#endif

#ifdef EXAR_IO
# define DEBUG_IO(...) printk(__VA_ARGS__)
#else
# define DEBUG_IO(...)
#endif



/**********************************************
*   Constants
*
***********************************************/

#define SFPGA_CPCR            (SFPGA_BASE_ADDR + 0x0010)  /* Console port control register */
#define CPCR_ENABLE_EXPANDER    0x100

#define IRQ_EXAR               6           /* GPIO_SUPFPGA_IRQ */
#define INT_OBSERVE_REG        0xB02008F8  /* Bit 20 of this register holds the Interrupt status of gpios */
#define GPIO_OBSERVE_INT       (1 << 20)

#define GPIO_P1_14_ADDR        0xB0200A78  /* GPIO port 1, bit 14 control register */
#define GPIO_P1_INTR_STATS     0xB020081C  /* GPIO port 1 interupt status */
#define GOT_INTR_PIN_14        (1 << 14)

/* Values for the GOIO bit control register */
#define GPIO_INTR_DISABLE      0x00000003
#define GPIO_INTR_RISE_EDGE    0x00000006
#define GPIO_INTR_FALL_EDGE    0x0000000A

/* Addresses for interrupt control registers in the FPGA */
#define SFPGA_EIER4           (SFPGA_BASE_ADDR + 0x3010)  /* Event Interrupt Enable Reg. 4   */
#define SFPGA_EITR4           (SFPGA_BASE_ADDR + 0x3110)  /* Event Interrupt Type Reg. 4     */
#define SFPGA_EIPR4           (SFPGA_BASE_ADDR + 0x3210)  /* Event Interrupt Polarity Reg. 4 */
#define SFPGA_EISR4           (SFPGA_BASE_ADDR + 0x3310)  /* Event Interrupt Status Reg. 4   */

/* The signal control bits are stored in consecutive 32-bit registers, hence mod 32 */
#define SFPGA_SIGNAL_BIT_RX_FULL   0x0002000  /* Signal 141 (0x0002000) */
#define SFPGA_SIGNAL_BIT_TX_EMPTY  0x0004000  /* Signal 142 (0x0004000) */
#define SFPGA_SIGNAL_BIT_RX_EMPTY  0x0008000  /* Signal 143 (0x0008000) */
#define SFPGA_EXPANDER_BITS    (SFPGA_SIGNAL_BIT_RX_FULL | SFPGA_SIGNAL_BIT_TX_EMPTY | SFPGA_SIGNAL_BIT_RX_EMPTY)

#define PASS_LIMIT          255
#define ENABLE              1
#define CLEAR               0
#define FIFO_DEPTH          64

/**********************************************
*   Macros
*
***********************************************/

#define SET_BIT(   addr,bitmask)   writeByte( addr, (readByte( addr) |   (bitmask) ))
#define CLR_BIT(   addr,bitmask)   writeByte( addr, (readByte( addr) & (~(bitmask))))
#define SET_BIT_16(addr,bitmask)   writeShort(addr, (readShort(addr) |   (bitmask) ))
#define CLR_BIT_16(addr,bitmask)   writeShort(addr, (readShort(addr) & (~(bitmask))))
#define SET_BIT_32(addr,bitmask)   writeLong( addr, (readLong( addr) |   (bitmask) ))
#define CLR_BIT_32(addr,bitmask)   writeLong( addr, (readLong( addr) & (~(bitmask))))

#define writeShort(addr, val) writeLong(addr, (unsigned long)val)
#define writeByte( addr, val) writeLong(addr, (unsigned long)val)
#define readShort(addr) (unsigned short)(readLong(addr) & 0xFFFF)
#define readByte( addr) (unsigned char)( readLong(addr) & 0xFF)

extern void register_irq_callback (
	void	(*irq_callback) (int irq, void *data),
	void	*data,
	int		driver_id);

/* Functions Prototype */
static void expander_isa_init_ports(void);
static int  receive_chars(struct uart_port *port);
static int  transmit_chars(struct uart_port *port);
static void expander_stop_tx(struct uart_port *port);

/* Global variables */
static struct uart_driver expander_reg;
static struct uart_port expander_ports;
static struct platform_device *expander_isa_devs;
static int transmit_enabled = 0;

#if SUPPORT_EXPANDER_UART_TASKLET
struct tasklet_struct expander_tasklet;
#endif

/* Functions to modify SFPGA registers
*  The SFPGA registers are composed of two 16-bit words in big-endian order.
*  In order to control the LSW of the 32-bit registers, we must add two to
*  access the correct word.
*/
static unsigned long readLong(unsigned long addr) {
	unsigned long val = (*((volatile unsigned short *)(addr    )));
	val = (val << 16) | (*((volatile unsigned short *)(addr + 2)));
	return val;
}

static void writeLong(unsigned long addr, unsigned long val) {
	*((volatile unsigned short *)(addr    )) = (unsigned short)((val >> 16)&0xFFFF);
	*((volatile unsigned short *)(addr + 2)) = (unsigned short)((val      )&0xFFFF);
}

/* read and write 32 bit values to SP memory or registers */
inline void sp_reg_write(unsigned long addr, unsigned long val) {
	*((volatile uint32_t*)addr) = val;
}

inline unsigned long sp_reg_read(unsigned long addr) {
	return ((*(volatile uint32_t*)addr));
}

/*****************************************************************************
* Reset UART RX, TX FIFO
******************************************************************************/
void reset_fifo(void) {

	/*No need to add an interval between RESET and OUT of RESET*/
	writeByte(EXP_FIFO_REG, (FIFO_RESET_RX|FIFO_RESET_TX));
	writeByte(EXP_FIFO_REG, 0);
}

/******************************************************************************
*  Interrupt Functions
******************************************************************************/

#ifdef ENABLE_EXAR_INTERRUPT
int exar_id = 4;
volatile unsigned char g_ier = 0;
#if SUPPORT_EXPANDER_UART_TASKLET
static void
expander_task(unsigned long privdata)
{
	int pass_counter = 0;
	int char_counter = 0;
	struct uart_port *port = (struct uart_port*)privdata;
	do {
		char_counter = 0;
		spin_lock(&port->lock);
		char_counter += receive_chars(port);
		char_counter += transmit_chars(port);
		spin_unlock(&port->lock);
	} while ((char_counter != 0) && (pass_counter++ <= PASS_LIMIT));

	/* If there is any kind of SFPGA interrupt pending, try to read the fifo
	 *  again. We must do this because the SFPGA emits level-based interupts,
	 * but the VSC452 can only detect edge transitions. If some other event
	 * causes an interrupt, it is possible that the VSC452 may miss the pending
	 * event for a character awaiting read from the uart. In this case, the
	 * fifo will fill, and we will lose characters. */
	if(readLong(SFPGA_EISR4) & SFPGA_EXPANDER_BITS) {
		writeLong(SFPGA_EISR4, SFPGA_EXPANDER_BITS);
		tasklet_schedule(&expander_tasklet);
	}
}
#else
static void
expander_task(struct uart_port *port)
{
	int pass_counter = 0;
	int char_counter = 0;

	do {
		char_counter = 0;
		spin_lock(&port->lock);
		char_counter += receive_chars(port);
		char_counter += transmit_chars(port);
		spin_unlock(&port->lock);
	} while ((char_counter != 0) && (pass_counter++ <= PASS_LIMIT));

}

#endif

irqreturn_t
expander_process_handler(struct uart_port *port) {
	DEBUG_INTR((KERN_ALERT "IP: ------ start ----------\n"));
	DEBUG_IO(KERN_ALERT "[irq--%X]\n", (unsigned int)readLong(SFPGA_EISR4));
	#if SUPPORT_EXPANDER_UART_TASKLET
	tasklet_schedule(&expander_tasklet);
	#else
	expander_task(port);
	#endif

	return IRQ_HANDLED;
}


static
irqreturn_t expander_intr_handler(int this_irq, void *dev_id)
{
	irqreturn_t retval = IRQ_NONE;

	/* Serial port information used by upper layers */
	struct uart_port *up = &expander_ports;

	/* Check if this was caused by the expander uart */
	if (readLong(SFPGA_EISR4) & SFPGA_EXPANDER_BITS) {
		writeLong(SFPGA_EISR4, SFPGA_EXPANDER_BITS);

		/* Service pending expander interrupt */
		if(IRQ_HANDLED == expander_process_handler(up)) {
			retval = IRQ_HANDLED;
		}
	}
	return retval;
}
#endif


/******************************************************************************
*  Polling Functions
******************************************************************************/

#ifdef ENABLE_EXAR_POLLING

static int start_kthread( struct uart_port *port);
static int stop_kthread(void);
static int expander_poll(struct uart_port *port);

static volatile int kthread_terminate = 0 ;
static pid_t pid;


static int
start_kthread( struct uart_port *port )
{
	DEBUG_PRINT(("Kernel thread for Expander UART \n"));

	kthread_terminate = 0;

	pid = kernel_thread((int(*)(void *))expander_poll, port, 0);
	if (pid < 0) {
		printk(KERN_ERR " Expander UART Fork Failed, Errno %d\n", -pid);
		return pid;
	}
	return 0;
}

static int
stop_kthread(void)
{
	kthread_terminate = 1;
	return 0;
}

static int
expander_poll(struct uart_port *port)
{
	daemonize("ExpanderThread");

	/* This loop shall  call either 'read_FIFO' (WRITE_CONSOLE) OR wakes up  the queue in 'expander_read_fifo' */
	while(kthread_terminate == 0) {
		receive_chars(port);
		transmit_chars(port);
		msleep(1);
	}
	return 0 ;
}
#endif


/******************************************************************************
*  General Functions
******************************************************************************/

static int
receive_chars(struct uart_port *port)
{
	struct  tty_struct *tty;
	int count = 0;
	char ch;
	int rx_reg;
	unsigned int continue_receive = 1;
	int status_reg = 0;
	int char_accepted = 0;


	if(port == NULL || port->state == NULL) {
		return 0;
	}
	tty = port->state->port.tty;

	DEBUG_INTR ((KERN_ALERT ">> RX : start : LCR :%x  \n", readByte((EXAR_BASE_ADDR + LSR_REG)) ));

	status_reg = readShort(EXP_FIFO_REG);
	while (((status_reg & FIFO_RX_EMPTY) == 0) && continue_receive) {
		rx_reg = readShort(EXP_RX_REG);
		ch = rx_reg & 0xFF;

		count++;
		port->icount.rx++;
		if (status_reg & FIFO_RX_FULL) {
			/*Since there is no flow control between SP and SFPGA, it's expected to lose characters when RX FIFO is full.*/
			printk(KERN_ERR"Expander UART FIFO FULL (0x%x)\n",(unsigned int)rx_reg);
		}

		/*tty_insert_flip_char() allocates the new buffer and returns the result
		  uart_insert_char doesn't return the result. So tty_insert_flip_char() is used.*/
		char_accepted=tty_insert_flip_char(tty,ch,TTY_NORMAL);
		if (char_accepted == 0) {
			/*No space in kernel buffer. Stop receiving.*/
			continue_receive = 0;
			printk(KERN_ERR"Expander UART: no space in kernel buffer.\n");
		}
		if (NULL != tty->buf.tail) {
			if (tty->buf.tail->used >= tty->buf.tail->size) {
				/*The last available byte is used. There is no available space in the current buffer.
				  Stop receiving and call tty_flip_buffer_push() after existing while() loop.*/
				continue_receive = 0;
			}
		}

		DEBUG_IO(KERN_ALERT "[r%02X:%c]\n",(unsigned int)ch,ch);
		status_reg = readShort(EXP_FIFO_REG);
	}

	if(count > 0) {
		spin_unlock(&port->lock);
		tty_flip_buffer_push(tty);
		spin_lock(&port->lock);
	}

	DEBUG_INTR ((KERN_ALERT "<< RX : END : LCR :%x  \n", readByte((EXAR_BASE_ADDR + LSR_REG)) ));
	return count;
}


static int
transmit_chars(struct uart_port *port)
{
	struct circ_buf  *xmit;
	int count = 0;

	if((port == NULL) || (port->state == NULL) || (transmit_enabled == 0)) {
		return 0;
	}
	xmit = &port->state->xmit;

	DEBUG_INTR ((KERN_ALERT "   TX : start : LCR :%x  \n",  readByte(EXP_LSR_REG) ));

	/* Transmit a single character */
	if (port->x_char && !(readByte(EXP_FIFO_REG) & FIFO_TX_FULL)) {
		DEBUG_INTR ((KERN_ALERT "       TX : port->x_char  \n" ));
		writeByte(EXP_TX_REG, port->x_char);
		DEBUG_IO(KERN_ALERT "[T%02X:%c]\n",(unsigned int)port->x_char,port->x_char);
		port->icount.tx++;
		port->x_char = 0;
		count++;
	}

	/* Check If Buffer TX buffer empty and return if empty  */
	else if (uart_circ_empty(xmit) || uart_tx_stopped(port)) {
		DEBUG_INTR ((KERN_ALERT "       TX : uart_circ_empty  ||  uart_tx_stopped \n" ));
		expander_stop_tx(port);
	}

	/* Transmit all characters in port FIFO */
	else {
		int fifosize;
		fifosize = port->fifosize;
		count=0;
		while( (count < fifosize) && !uart_circ_empty(xmit) && !(readByte(EXP_FIFO_REG) & FIFO_TX_FULL)) {
			DEBUG_INTR ((KERN_ALERT "       TX : %c \n", xmit->buf[xmit->tail]));
			writeByte(EXP_TX_REG, xmit->buf[xmit->tail]);
			DEBUG_IO(KERN_ALERT "[t%02X:%c]\n",(unsigned int)xmit->buf[xmit->tail],xmit->buf[xmit->tail]);
			xmit->tail = (xmit->tail + 1) & (UART_XMIT_SIZE - 1);
			port->icount.tx++;
			count++;
		}

		if (uart_circ_chars_pending(xmit) < WAKEUP_CHARS) {
			uart_write_wakeup(port);
			DEBUG_INTR ((KERN_ALERT "       TX : uart_write_wakeup \n "));
		}

		if (uart_circ_empty(xmit)) {
			DEBUG_INTR ((KERN_ALERT "       TX : uart_circ_empty  \n"));
			expander_stop_tx(port);
		}
	}

	DEBUG_INTR((KERN_ALERT "   TX : end  : LCR :%x  \n",  readByte(EXP_LSR_REG) ));
	return count;
}

static void
expander_stop_tx(struct uart_port *port) {
#ifdef ENABLE_EXAR_INTERRUPT
	CLR_BIT_32(SFPGA_EIER4, SFPGA_SIGNAL_BIT_TX_EMPTY);     /* Disable TX interrupt */
	transmit_enabled = 0;
#endif

	DEBUG_PRINT((KERN_ERR "expander_stop_tx\n"));
	return;
}

static void
expander_stop_rx (struct uart_port *port) {
#ifdef ENABLE_EXAR_INTERRUPT
	CLR_BIT_32(SFPGA_EIER4, (SFPGA_SIGNAL_BIT_RX_FULL|SFPGA_SIGNAL_BIT_RX_EMPTY));     /* Disable RX interrupt */
#endif

	DEBUG_PRINT((KERN_ERR "expander_stop_rx \n"));
	return;
}

static void
expander_start_tx(struct uart_port *port) {
	transmit_enabled = 1;
#ifdef ENABLE_EXAR_INTERRUPT
	SET_BIT_32(SFPGA_EIER4, SFPGA_SIGNAL_BIT_TX_EMPTY);     /* Enable TX interrupt */
#endif

	/*We need to transmit data to SFPGA here.
	  When data are sent out by SFPGA, tx_empty interrupt is triggered (edge triggerred).*/
	transmit_chars(port);

	DEBUG_PRINT((KERN_ERR "expander_start_tx \n"));
	return;
}


static int
expander_startup(struct uart_port *port) {

#ifdef ENABLE_EXAR_INTERRUPT
	/*Reset RX,TX FIFO*/
	reset_fifo();

	/* Configure GPIO P1_14 as input, issue interrupt on falling edge */
	sp_reg_write(GPIO_P1_14_ADDR, GPIO_INTR_FALL_EDGE);

	/* Enable SFPGA expander interrupts */
	CLR_BIT_32(SFPGA_EITR4, SFPGA_EXPANDER_BITS);        /* Only trigger on one edge */
	CLR_BIT_32(SFPGA_EIPR4, SFPGA_SIGNAL_BIT_RX_FULL);   /* Trigger on rising edge */
	CLR_BIT_32(SFPGA_EIPR4, SFPGA_SIGNAL_BIT_TX_EMPTY);  /* Trigger on rising edge */
	SET_BIT_32(SFPGA_EIPR4, SFPGA_SIGNAL_BIT_RX_EMPTY);  /* Trigger on falling edge */
	writeLong(SFPGA_EISR4, SFPGA_EXPANDER_BITS);        /* Clear pending interrupt */
	SET_BIT_32(SFPGA_EIER4, (SFPGA_SIGNAL_BIT_RX_EMPTY | SFPGA_SIGNAL_BIT_RX_FULL));  /* Enable new RX interrupts */
	CLR_BIT_32(SFPGA_EIER4, SFPGA_SIGNAL_BIT_TX_EMPTY);     /* Disable TX interrupt */

#endif

#ifdef ENABLE_EXAR_POLLING
	writeByte(EXP_LCR_REG, 0xBF);
	writeByte(EXP_EFR_REG, 0x00);
	writeByte(EXP_IER_REG, 0);
	writeByte(EXP_LCR_REG, 0);

	/*
	* Clear the FIFO buffers and disable them.
	* (they will be enabled in set_termios())
	*/
	writeByte(EXP_FCR_REG,  TX_RX_FIFO_ENABLE);
	writeByte(EXP_FCR_REG, (TX_RX_FIFO_ENABLE | TX_FIFO_RESET | RX_FIFO_RESET));
	writeByte(EXP_FCR_REG, 0);

	/*
	 * Now, initialize the UART
	 */
	writeByte(EXP_LCR_REG, (WORD_LEN_8 | STOP_BIT_LEN_1 | NO_PARITY));

	start_kthread(port);
#endif

	/* Convert carriage return/new line usage */
	port->state->port.tty->termios->c_iflag &= ~ICRNL;
	port->state->port.tty->termios->c_oflag &= ~ONLCR;

	/* Disable local echo, this prevents a feedback loop since the expander echoes too */
	port->state->port.tty->termios->c_lflag &= ~(ICANON | ECHO | ECHOCTL);

	/* Enable path for expander data */
	SET_BIT_16(SFPGA_CPCR, CPCR_ENABLE_EXPANDER);

	transmit_enabled = 0;
	DEBUG_PRINT((KERN_ALERT "expander_startup\n"));
	return 0;
}

static void
expander_shutdown(struct uart_port *port) {
#ifdef ENABLE_EXAR_INTERRUPT
	CLR_BIT_32(SFPGA_EIER4, SFPGA_EXPANDER_BITS);        /* Disable RX, TX interrupts */
#endif

#ifdef ENABLE_EXAR_POLLING
	/*
	* Disable break condition and FIFOs
	*/
	writeByte(EXP_FCR_REG, readByte(EXP_LCR_REG) & ~ LCR_SBC);
	writeByte(EXP_FCR_REG, (TX_RX_FIFO_ENABLE | TX_FIFO_RESET | RX_FIFO_RESET));
	writeByte(EXP_FCR_REG, 0);

	/*
	 * Read data port to reset things, and then unlink from
	 * the IRQ chain.
	 */
	readByte(EXP_RX_REG);

	stop_kthread();
#endif
	/* Disable path for expander data */
	CLR_BIT_16(SFPGA_CPCR, CPCR_ENABLE_EXPANDER);

	transmit_enabled = 0;
	DEBUG_PRINT((KERN_ERR "expander_uart_shutdown\n"));
	return;
}

static void
expander_set_termios(struct uart_port *port, struct ktermios *termios, struct ktermios *old) {
	struct tty_struct *tty = port->state->port.tty;
	unsigned int baud = 0;

	DEBUG_PRINT((KERN_ERR "expander_set_termios\n"));

	baud = tty_get_baud_rate(tty);
	DEBUG_PRINT(("baud rate %d\n", baud));
	return;
}

/*
*   Unused standard SERIAL UART operations (ie)  expander_pops Functions defintions
*/

static unsigned int
expander_tx_empty(struct uart_port *port) {
	DEBUG_PRINT((KERN_ERR "expander_tx_empty\n"));
	return TIOCSER_TEMT;
}

static unsigned int
expander_get_mctrl(struct uart_port *port) {
	DEBUG_PRINT((KERN_ERR "expander_get_mctrl\n"));
	return 0;
}

static  void
expander_set_mctrl(struct uart_port *port, unsigned int mctrl) {
	DEBUG_PRINT((KERN_ERR "expander_set_mctrl\n"));
	return;
}

static void
expander_enable_ms(struct uart_port *port) {
	DEBUG_PRINT((KERN_ERR "expander_enable_ms\n"));
	return;
}

static void
expander_break_ctl(struct uart_port *port, int break_state) {
	DEBUG_PRINT((KERN_ERR "expander_break_ctl\n"));
	return;
}

static void
expander_pm(struct uart_port *port, unsigned int state, unsigned int oldstate) {
	DEBUG_PRINT((KERN_ERR "expander_pm\n"));
	return;
}

static void
expander_release_port(struct uart_port *port) {
	DEBUG_PRINT((KERN_ERR "expander_release_port\n"));
	return;
}

static int
expander_request_port(struct uart_port *port) {
	DEBUG_PRINT((KERN_ERR "expander_request_port\n"));
	return 0;
}

static void
expander_config_port(struct uart_port *port, int flags) {
	DEBUG_PRINT((KERN_ERR "expander_config_port \n"));
	return;
}

static int
expander_verify_port(struct uart_port *port, struct serial_struct *ser) {
	DEBUG_PRINT((KERN_ERR "expander_verify_port\n"));
	return -EINVAL;
}

static const char *
expander_type(struct uart_port *port) {
	DEBUG_PRINT((KERN_ERR "expander_type\n"));
	return "Exaruart";
}

static struct uart_ops expander_pops = {
	.tx_empty     = expander_tx_empty,
	.set_mctrl    = expander_set_mctrl,
	.get_mctrl    = expander_get_mctrl,
	.stop_tx      = expander_stop_tx,
	.start_tx     = expander_start_tx,
	.stop_rx      = expander_stop_rx,
	.enable_ms    = expander_enable_ms,
	.break_ctl    = expander_break_ctl,
	.startup      = expander_startup,
	.shutdown     = expander_shutdown,
	.set_termios  = expander_set_termios,
	.pm           = expander_pm,
	.type         = expander_type,
	.release_port = expander_release_port,
	.request_port = expander_request_port,
	.config_port  = expander_config_port,
	.verify_port  = expander_verify_port,
};


#ifdef CONFIG_EXAR_16M770_CONSOLE

static void expander_console_putchar(struct uart_port *port, int ch) {
	DEBUG_PRINT((KERN_ERR "expander_console_putchar\n"));
}


/*
 *  Print a string to the serial port trying not to disturb
 *  any possible real use of the port...
 *
 *  The console_lock must be held when we get here.
 */
static void
expander_console_write(struct console *co, const char *s, unsigned int count) {
	struct uart_port *port = &expander_ports;

	touch_nmi_watchdog();
	spin_lock(&port->lock);
	DEBUG_PRINT((KERN_ERR "expander_console_write\n"));
	uart_console_write(port, s, count, expander_console_putchar);
	spin_unlock(port->lock);
}

static int
expander_console_setup(struct console *co, char *options) {
	int baud = 115200;
	int bits = 8;
	int parity = 'n';
	int flow = 'n';

	if (options) {
		uart_parse_options(options, &baud, &parity, &bits, &flow);
	}

	return uart_set_options(&expander_ports, co, baud, parity, bits, flow);
}

static struct console expander_console = {
	.name   = "ttyXR",
	.write  = expander_console_write,
	.device = uart_console_device,
	.setup  = expander_console_setup,
	.flags  = CON_PRINTBUFFER,
	.index  = -1,
	.data   = &expander_reg,
};

static int __init
expander_console_init(void) {
	expander_isa_init_ports();
	register_console(&expander_console);
	return 0;
}

	#define EXAR_CONSOLE      &expander_console
#else
	#define EXAR_CONSOLE      NULL
#endif


static struct uart_driver expander_reg = {
	.owner       = THIS_MODULE,
	.driver_name = "XR16M",
	.dev_name    = "ttyXR",
	.major       = TTY_MAJOR,
	.minor       = 68,
	.nr          = UART_NR,
	.cons        = EXAR_CONSOLE,
};


static void
expander_isa_init_ports(void) {
	struct uart_port *port = (struct uart_port *)&expander_ports;

	port->line = 0;
	spin_lock_init(&port->lock);

	port->ops      = &expander_pops;
	port->type     = 1000; // Just so we're not "unknown"
	port->fifosize = FIFO_DEPTH;
	port->uartclk  = 115200 * 16; // Dummy to avoid getting dropped in case serial_core cares
	port->flags    = 0;
	port->hub6     = 0;
	port->mapbase  = EXP_UART_BASE_ADDR;
	port->membase  = (char*)(EXP_UART_BASE_ADDR);
	port->iotype   = UPIO_MEM32;
	port->regshift = 0;
}

static int __init
expander_register_ports(struct uart_driver *drv, struct device *dev) {
	int ret = 0;
	#if SUPPORT_EXPANDER_UART_TASKLET
	struct uart_port *port = (struct uart_port *)&expander_ports;
	tasklet_init(&expander_tasklet, expander_task, (unsigned long)port);
	#endif
	DEBUG_PRINT((KERN_ERR "expander_register_ports\n"));

#ifdef CONFIG_EXAR_16M770_CONSOLE
	expander_console_init();
#else
	expander_isa_init_ports();
#endif

	/*
	*  uart_add_one_port - attach a driver-defined port structure
	*  @drv: pointer to the uart low level driver structure for this port >>> "tty"  structure which is registered as serial
	*  @port: uart port structure to use for this port.                   >>> "uart_port" structure which has driver specfic fnctions like irq, lock etc
	*
	*  This allows the driver to register its own uart_port structure
	*  with the core driver.  The main purpose is to allow the low
	*  level uart drivers to expand uart_port, rather than having yet
	*  more levels of structures.
	*/
	uart_add_one_port(drv, &expander_ports);
	/*Reset RX,TX FIFO*/
	reset_fifo();

#ifdef ENABLE_EXAR_INTERRUPT
	#ifdef SUPPORT_EXPANDER_UART_IRQ  //This is not supported.
	/*
	*  Register IRQ
	*/
	ret = request_irq(IRQ_EXAR, expander_intr_handler, IRQF_SHARED, "expander_uart", &exar_id);
	if (ret < 0) {
		printk(KERN_ERR "Retval : %d\n",ret);
		printk(KERN_ERR "Reqt for irq %d failed\n", IRQ_EXAR);
		ret= -ENODEV;
	}
	#else
	register_irq_callback((void *)expander_intr_handler,&exar_id,EXPANDER_UART_ID);
	#endif
	/* Configure GPIO P1_14 as input, issue interrupt on falling edge */
	sp_reg_write(GPIO_P1_14_ADDR, GPIO_INTR_FALL_EDGE);
#endif
	return ret;
}


static int __init
expander_init(void) {
	int ret = -ENODEV;
	printk (KERN_ALERT "Loading Expander UART Driver Version %s\n", VERSION_STRING);

	/* Register tty as serial interface */
	if(uart_register_driver(&expander_reg) == 0) {
		ret = expander_register_ports(&expander_reg, &expander_isa_devs->dev);
	}
	return ret;
}

static void __exit
expander_exit(void) {
	printk (KERN_ALERT "Unloading Expander Uart Driver \n");
	uart_remove_one_port(&expander_reg, &expander_ports);
	uart_unregister_driver(&expander_reg);

#ifdef ENABLE_EXAR_INTERRUPT
	#ifdef SUPPORT_EXPANDER_UART_IRQ  //This is not supported.
	/*  Free IRQ */
	free_irq(IRQ_EXAR, &exar_id);
	#else
	register_irq_callback(NULL,NULL,EXPANDER_UART_ID);
	#endif
	/* Disable SFPGA expander interrupts */
	CLR_BIT_32(SFPGA_EIER4, SFPGA_EXPANDER_BITS);  /* Disable new interrupts */
	writeLong(SFPGA_EISR4, SFPGA_EXPANDER_BITS);  /* Clear pending interrupt */
#endif
}

/**
 *      expander_suspend_port - suspend one serial port
 *      @line:  serial line number
 *
 *      Suspend one serial port.
 */
void
expander_suspend_port(int line) {
	uart_suspend_port(&expander_reg, &expander_ports);
}

/**
 *      expander_resume_port - resume one serial port
 *      @line:  serial line number
 *      @level: the level of port resumption, as per uart_resume_port
 *
 *      Resume one serial port.
 */
void expander_resume_port(int line) {
	uart_resume_port(&expander_reg, &expander_ports);
}

module_init(expander_init);
module_exit(expander_exit);

EXPORT_SYMBOL(expander_suspend_port);
EXPORT_SYMBOL(expander_resume_port);

MODULE_AUTHOR ("American Megatrends Inc");
MODULE_DESCRIPTION ("Expander UART driver Module with Polling Mode Operation ");
MODULE_LICENSE ("GPL");
